package workshop.core;

import java.util.*;
import workshop.utils.*;
import workshop.core.exceptions.*;

/**
 * @author Michele
 * @author Beniamino
 * @version 1.10
 *
 * Classe JobsSupervisor. Implementa un thread che continua a pescare un
 * lavoro dalla lista dei job in attesa. Se la lista  vuota, alla prima get
 * sulla lista si aspetta un certo tempo, prima del tentativo successivo. La
 * funzione che recupera un lavoro da assegnare, ricerca all'interno della lista
 * se un lavoro non ha abbastanza risorse a disposizione per essere svolto.
 *
 * Se un job viene saltato troppe volte perch richiede troppe risorse, ad un
 * certo punto, ottiene un valore di priorit massimo che indica che deve essere
 * svolto prima di tutti gli altri. In pratica, il supervisor non controlla pi
 * i job successivi in coda.
 */

public final class JobsSupervisor implements Runnable
{
   /**
    * Il costruttore
    */

   private JobsSupervisor()
   {
      selfThread = new Thread(this, "JobsSupervisor");
      selfThread.start();
   }
   
   /**
    * Aggiunge un job alla coda dei lavori
    *
    * @param j Il job da aggiungere in coda
    */

   public void addJob(Job j)
   {
      jobQueue.add((Job) j);
   }

   /**
    * Recupera un job dalla coda dei lavori
    *
    * @return Il job da assegnare, o null se la coda  vuota o non ci sono job
    *         assegnabili a causa delle troppe risorse che richiedono
    */

   private Job takeJob()
   {
      ListIterator<Job> jobs = jobQueue.listIterator();
      
      /* Comincio la visita all'interno della coda */

      while (jobs.hasNext())
      {
         jobPending = null;
         
         Job j = jobs.next();

         /* Se il job corrente aveva abbastanza risorse, lo rimuovo dalla coda
          * e lo uso come valore di ritorno */

         if ( enoughTool(j) )
         {
            reserveTools(j);
            return jobQueue.remove(jobQueue.indexOf(j));
         }

         j.increaseTimesSkipped();
   
         /* Se il job era stato saltato troppe volte aspetto finch le risorse
          * sono state liberate */

         if ( j.getTimesSkipped() >= ConfRegister.MAX_JOB_SKIPS )         
         {
   			waitingToolStart = System.currentTimeMillis();
			   waitForToolTimeDone += 1;

            synchronized (this)
            {
               try
               {
                  jobPending = j;
                  
                  Logger.debug("Attendo che i laborer liberino le risorse necessarie!!!\n", this.selfThread, false);
                  this.wait();
                  Logger.debug("Ho trovato gli attrezzi necessari! Recupero il job pendente!\n", this.selfThread, false);
               }
               catch (InterruptedException ie)
               {
               }
            }
            
            waitingToolCurrent = System.currentTimeMillis() - waitingToolStart;
         
            synchronized (Statistics.lock)
            {
               waitingToolAverage = (waitingToolAverage * waitForToolTimeDone  + waitingToolCurrent) / (waitForToolTimeDone + 1);
   
               if (waitingToolCurrent > waitingToolMax)
                  waitingToolMax = waitingToolCurrent;
            }

            reserveTools(j);

            return jobQueue.remove(jobQueue.indexOf(j));
         }
      }

      return null;
   }
   
   /**
    * Aggiunge un operaio nella coda degli operai inattivi, in attesa di job
    *
    * @param laborer L'operaio da inserire in coda
    */

   public void addLaborer(Laborer laborer)
   {
      laborerQueue.add(laborer);
      
      if (jobPending != null && enoughTool(jobPending))
      {
         synchronized (this)
         {
            Logger.debug("Un laborer (al piu' l'ultimo in RUN) ha liberato risorse per svolgere il job " + jobPending.getID() + ". Sveglio Supervisor!\n", true);
            this.notify();
         }
      }
      else
      {
         synchronized (laborerQueue)
         {
            laborerQueue.notify();
         }
      }
   }
   
   /**
    * Preleva un laborer dalla coda degli operai, per assegnargli il lavoro
    * appena estratto. Se la coda degli operai  vuota, il thread resta in
    * attesa, finch un laborer non si annuncia invocando il metodo addLaborer.
    *
    * @return L'operaio estratto dalla coda
    */

   private Laborer takeLaborer()
   {
      Laborer l = null;
      
      synchronized (laborerQueue)
      {
         if (laborerQueue.isEmpty())
         {
            try
            {
               Logger.debug("Attendo un laborer libero!!!\n", false);
               laborerQueue.wait();
            }
            catch (InterruptedException ie)
            {
            }
         }
   
         l = this.laborerQueue.remove();
      }
      
      return l;
   }
   
   /**
    * Il metodo RUN per il thread
    */

   public void run()
   {
      while (true)
      {
         if (status == PAUSE)
         {
            try
            {
               synchronized (this)
               {
                  this.wait();
               }
            }
            catch (InterruptedException ie)
            {
            }
         }
         
         /* Recupera un job dalla lista */
         Job job = takeJob();

         if (job == null)
         {
         	
            try
            {
               /* Eseguita solo quando la coda  vuota o non ha lavori,
                * con risorse insufficienti, da eseguire immediatamente */

               this.selfThread.sleep(ConfRegister.PAUSE_TIME_ON_NULL_JOB);
            }
            catch (InterruptedException ie)
            {
            }
         }
         else
         {
            Logger.debug("Non dovrei essere in attesa che i laborer finiscano!!\n", this.selfThread, false);
            
            /* Se ho recuperato un job valido, leggo un operaio dalla coda e gli
             * assegno il lavoro appena estratto. TakeLaborer pu fermare il thread
             * in uno stato di wait. */

            Laborer lab = takeLaborer();

            /* Assegno il lavoro all'operaio, forzandogli l'uscita dal suo stato di wait_job */

            lab.assignJob(job);
         }
      }
   }
   
   
   private boolean enoughTool(Job j)
   {
      /* Per ogni tool che il job richiede, controllo se ho abbastanza copie
       * per assegnare il lavoro */
   
      for ( Tool t : j.getToolsNeeded() )
      {
         String toolkey = t.getName();

         /* Se non ho abbastanza risorse, abbasso il flag e provo a
          * controllare se il prossimo job ha risorse a sufficienza per
          * essere eseguito */

         if ( Toolbox.tools().selectTool(toolkey).getCopies() < t.getCopies() )
         {
            return false;
         }
      }

      return true;
   }
   
   private void reserveTools(Job j)
   {
      for (Tool t : j.getToolsNeeded())
      {
         try
         {
            Toolbox.tools().selectTool(t.getName()).decrease(t.getCopies());
         }
         catch(UnderflowException e)
         {
            Logger.error(this.selfThread.getName() + ": ERRORE NELLA PRENOTAZIONE DELLE RISORSE!!", e);
         }
      }
   }
   
   /**
    * Accessor per l'unica istanza della classe
    *
    * @return L'istanza della classe JobsSupervisor corrente
    */

   public static JobsSupervisor self()
   {
      return DEFAULT_SUPERVISOR;
   }

   /**
    * @return Il tempo medio di attesa per gli strumenti
    */

   public static synchronized long getWaitingToolAverageTime()
   {
      return waitingToolAverage;
   }

   /**
    * @return Il tempo massimo di attesa per gli strumenti
    *         
    */
   
   public static synchronized long getWaitingJobMaxTime()
   {
      return waitingToolMax;
   }

   private Job jobPending = null;

   private static volatile LinkedList<Job>      jobQueue       = new LinkedList<Job>();
   private static volatile LinkedList<Laborer>  laborerQueue   = new LinkedList<Laborer>();

   private static JobsSupervisor DEFAULT_SUPERVISOR = new JobsSupervisor();
   
   private Thread selfThread = null;
   
   private volatile int status = RUN;
   
   
  
   private static volatile int   waitForToolTimeDone    = 0;
   private static volatile long waitingToolAverage = 0;
   private static volatile long waitingToolMax = 0;
   private static volatile long waitingToolStart,waitingToolCurrent;
   
   public static final int RUN    = 0;
   public static final int PAUSE  = 1;
   
   public synchronized void pause()
   {
      if (status == RUN)
         status = PAUSE;
   }
   
   public synchronized void resume()
   {
      if (status == PAUSE)
      {
         status = RUN;
         this.notify();
      }
   }
}