package workshop.core;

import java.util.StringTokenizer;

import workshop.core.exceptions.*;
import workshop.utils.*;

/**
 * @author Michele
 * @author Beniamino
 * @version 0.1
 * @see Toolbox
 *
 * Classe Tool. Implementa un insieme di risorse, con il relativo numero di copie
 * che, a seconda del contesto, indicano il numero di copie disponibili, prenotate
 * o di cui l'operaio ha bisogno.
 */

public class Tool
{  
   /**
    * Il costruttore
    *
    * @param name   Il nome dello strumento
    * @param copies Le copie disponibili/richieste del tool
    */

   public Tool(String name, int copies)
   {
      try
      {
         setName(name);
         setCopies(copies);
      }
      catch (IllegalArgumentException e)
      {
         Logger.error("illegal argument during creation of tool " + name, e);
      }
   }
   
   /**
    * Il costruttore
    *
    * @param name      Il nome dello strumento
    * @param copies    Le copie disponibili/richieste del tool
    * @param maxcopies Il massimo numero di copie disponibili per questo tool,
    *                  utilizzato nell'inizializzazione della Toolbox
    */

   public Tool(String name, int copies, int maxcopies)
   {
      try
      {
         setName(name);
         setCopies(copies);
         setMaxCopies(maxcopies);
      }
      catch (IllegalArgumentException e)
      {
         Logger.error("illegal argument during creation of tool " + name, e);
      }
   }

   /**
    * @return Il numero di copie disponibili attualmente
    */
    
   public int getCopies()
   {
      return uToolCopies;
   }
   
   /**
    * @return Il massimo numero di copie disponibili
    */
    
   public int getMaxCopies()
   {
      return maxCopies;
   }
   
   /**
    * @return Il nome dello strumento
    */
   
   public String getName()
   {
      return toolName;
   }

   /**
    * Setta il numero di copie dello strumento
    *
    * @param copies Il numero di copie
    */
   
   private void setCopies(int copies) throws IllegalArgumentException
   {
      if (copies < 1)
      {
         throw new IllegalArgumentException();
      }

      uToolCopies = copies;
   }
   
   /**
    * Setta il massimo numero di copie dello strumento
    *
    * @param copies Il numero di copie
    */
   
   private void setMaxCopies(int copies) throws IllegalArgumentException
   {
      if (copies < 1)
      {
         throw new IllegalArgumentException();
      }

      maxCopies = copies;
   }
   
   /**
    * Setta il nome dello strumento per l'instanza attuale
    *
    * @param name Il nome del tool
    */
   
   private void setName(String name) throws IllegalArgumentException
   {
      toolName = new String(name);
   }

   /**
    * Decrementa il numero di copie attuali per il tool corrente
    */

   public synchronized void decrease() throws UnderflowException
   {
      if (uToolCopies <= 0)
         throw new UnderflowException();
         
      uToolCopies--;
   }

   /**
    * Decrementa il numero di copie attuali per il tool corrente del valore passato
    *
    * @param number Il numero di copie da decrementare
    */

   public synchronized void decrease(int number) throws UnderflowException,IllegalArgumentException
   {
      if(number < 1)
         throw new IllegalArgumentException();
      
      if (uToolCopies < number)
         throw new UnderflowException();
         
      uToolCopies -= number;
   }

   /**
    * Incrementa il numero di copie attuali per il tool corrente
    */

   public synchronized void increase() throws OverflowException
   {
      if (uToolCopies >= maxCopies)
         throw new OverflowException();
                                      
      uToolCopies++;
   }
   
   /**
    * Aggiunge N copie dello strumento al numero di copie attuali
    *
    * @param number Il numero di copie da aggiungere
    */

   public synchronized void increase(int number) throws OverflowException,IllegalArgumentException
   {
      if(number < 1)
         throw new IllegalArgumentException();
      
      if ((uToolCopies + number) > maxCopies)
         throw new OverflowException();
         
      uToolCopies += number;
   }
   
   private           String   toolName       = null;
   private volatile  int      uToolCopies    = 0;
   private           int      maxCopies;
}