package workshop.core;

import java.io.*;
import java.util.*;

import workshop.core.exceptions.*;
import workshop.utils.ConfRegister;

/**
 * @author Michele
 * @author Beniamino
 * @version 0.1
 * @see Tool
 *
 * Classe Toolbox. Rappresenta la cassetta degli attrezzi, dove vengono riposte
 * le copie disponibili dei tools. Non  derivabile, quindi  dichiarata final.
 * Non voglio che qualcuno instanzi la classe, quindi l'unico costruttore 
 * dichiarato privato.
 */
 
public final class Toolbox
{
   /**
    * Il costruttore. Dichiarato privato perch c' una sola toolbox in tutta
    * l'officina.
    *
    * Alla prima chiamata alla classe, viene caricato il file di configurazione
    * e viene riempita la hashmap degli strumenti.
    */
    
   private Toolbox()
   {
      FileInputStream   fileInputStream  = null;
      String            filePath         = null;
      
      int ch;

      filePath = ConfRegister.getConfigurationFilePath();
      
      try 
      {
         fileInputStream = new FileInputStream(filePath);
         
         toolsBuffer = new StringBuffer();
         
         // Fill the StringBuffer reading from the file
         while ((ch = fileInputStream.read()) != -1)
         {
            toolsBuffer.append((char) ch);
         }
         
         // Fill the hashtable, loading Tools from the StringBuffer
         loadTools();
      }
      catch (FileNotFoundException fnfe)
      {
         Logger.error("configuration file has not been found.");
      }
      catch (IOException ioe)
      {
         Logger.error("IO exception during config file reading.", ioe);
      }
      catch (SyntaxException se)
      {
         Logger.error("configuration file contains syntax error.");
      }
      finally
      {
         try
         {
            fileInputStream.close();
         }
         catch (IOException ioex)
         {
            Logger.error("cannot close the config file.", ioex);
         }
      }
   }
   
   /**
    * @return Il numero di tools presenti
    */

   public int countTools()
   {
      return toolHashTable.size();
   }
   
   /**
    * @return Un Enumeration contenente i nomi di tutti gli strumenti caricati
    */
   
   public Enumeration getToolNames()
   {
      return toolHashTable.keys();
   }
   
   /**
    * Carica la HashTable a partire dalla immensa stringa passatagli, contenente
    * la configurazione appena caricata dal file.
    */
   private void loadTools() throws SyntaxException
   {
      StringTokenizer st = new StringTokenizer(toolsBuffer.toString());
      
      boolean uniqueCopyExists = false;

      if (st.countTokens() % 2 != 0)
      {
         throw new SyntaxException();
      }

      while (st.hasMoreTokens())
      {
         try
         {
            String toolkey = st.nextToken();
            int copies = Integer.parseInt(st.nextToken());
            
            if (copies == 1)
            {
               uniqueCopyExists = true;
            }

            Logger.debug("Toolkey is: " + toolkey + "\n");

            toolHashTable.put( (String) toolkey, (Tool) new Tool( toolkey, copies, copies) );
         }
         catch (Exception e)
         {
            Logger.error("Toolbox: ERRORE NEL CARICAMENTO DEL FILE DI CONFIGURAZIONE!!", e);
         }
      }
      
      if (!uniqueCopyExists)
      {
         Logger.warning("almost one tool must have a unique copy");
      }
   }
   
   /**
    * Seleziona un tool dalla HashMap
    *
    * @param  toolkey Il nome dello strumento, usato come chiave
    * @return L'insieme dei tools (vuoto o meno) all'interno della Toolbox (lo scaffale)
    */
   
   public Tool selectTool( String toolkey )
   {
      Tool t = (Tool) toolHashTable.get( toolkey );

      if ( t != null )
      {
         return t;
      }

      return null;
   }

   /**
    * Ritorna l'istanza della classe
    */

   public static Toolbox tools()
   {
      return DEFAULT_TOOLBOX;
   }

   private static Toolbox DEFAULT_TOOLBOX = new Toolbox();
   
   public static final int defaultMaxCopies  = 6;
   
   private Hashtable<String,Tool>   toolHashTable     = new Hashtable<String,Tool>();
   private StringBuffer             toolsBuffer       = null;
}